mod cli;
mod packet;
mod tm;
mod utils;

use std::{fs::File, time::Instant};

use clap::Parser;
use color_eyre::eyre::Result;
use log::{debug, info};
use packet::TimedMessage;
use skyward_mavlink::{
    lyra::MavMessage,
    mavlink::{self, MavConnection, MavHeader, MavlinkVersion},
};
use tm::read_high_rate_telemetry;

use crate::{cli::Cli, packet::PacketSequence, utils::print_stats};

fn main() -> Result<()> {
    color_eyre::install()?;
    let args: Cli = Cli::parse();

    let lvl = if args.verbose {
        log::LevelFilter::Debug
    } else {
        log::LevelFilter::Info
    };

    // initialize logger
    simple_logger::SimpleLogger::new()
        .with_level(lvl)
        .with_local_timestamps()
        .env()
        .init()?;

    let packets = get_packets(&args)?;
    debug!("collected {} packets", packets.len());

    // convert to packet sequence data structure
    let mut packets = PacketSequence::from(packets);

    // Setup mavlink connection
    let mut mavconn = mavlink::connect::<MavMessage>(&format!("udpout:127.0.0.1:{}", args.port))?;
    // Lyra dialect use Mavlink version 1
    mavconn.set_protocol_version(MavlinkVersion::V1);
    let _header = mavlink::MavHeader {
        system_id: args.system_id,
        component_id: 0,
        sequence: 0,
    };
    debug!("mavlink flushing on UDP port {}", args.port);

    // read from buffer and print to stdout
    let mut start: Option<Instant> = None;
    let mut times = Vec::new();

    // send packet if there is one
    while let Some(packet) = packets.wait_next() {
        if let Some(t) = start {
            times.push(t.elapsed().as_millis());
        }
        start = Some(Instant::now());
        mavconn.write_packet(packet.to_owned(), args.system_id)?;
        debug!(
            "sent packet ({}/{})",
            packets.sent_packets(),
            packets.total_packets()
        );
    }

    info!("sent {} packets", packets.sent_packets());
    print_stats(times);

    Ok(())
}

fn get_packets(args: &Cli) -> Result<Vec<TimedMessage>> {
    info!(
        "reading high rate main telemetry from file {}",
        args.high_rate_file_main.display()
    );
    info!(
        "reading high rate payload telemetry from file {}",
        args.high_rate_file_payload.display()
    );
    let hrtf_main = File::open(&args.high_rate_file_main)?;
    let hrtf_payload = File::open(&args.high_rate_file_payload)?;
    let mut packets = read_high_rate_telemetry(hrtf_main, hrtf_payload)?;
    packets.sort();
    debug!("collected {} packets", packets.len());
    Ok(packets)
}

trait MavLinkPort {
    fn write_packet(&mut self, packet: MavMessage, system_id: u8) -> Result<()>;
}

impl<T: MavConnection<MavMessage> + ?Sized> MavLinkPort for T {
    fn write_packet(&mut self, packet: MavMessage, system_id: u8) -> Result<()> {
        let header = MavHeader {
            system_id,
            ..Default::default()
        };
        self.send(&header, &packet)?;
        Ok(())
    }
}
