import os
from pathlib import Path

import click
import polars as pl

NAS_CALIBRATE_STATE = 1
MAIN_FLYING_FMM_STATE = 10
PAYLOAD_FLYING_FMM_STATE = 9
SECONDS_BEFORE_FLYING = 15


@click.command()
@click.argument("path", type=click.Path(exists=True))
@click.option("--output", type=click.Path(), default=".")
@click.option("--origin-main", "origin", flag_value="main", default=True)
@click.option("--origin-payload", "origin", flag_value="payload")
def main(path: Path, output: Path, origin: str):
    """
    Convert the logs found in the given path to a format compliant with ARPIST.
    """
    # now walk in the directory pointed by path the files: main_Boardcore_EventData.csv, main_Boardcore_NASState.csv, main_Boardcore_ReferenceValues.csv and save their path to variables
    nas_controller_status_path = None
    nas_state_path = None
    ada_state_path = None
    fmm_status_path = None
    gps_path = None

    for p, _, fn in os.walk(path):
        for f in fn:
            if origin == "main":
                if f == "main_Main_NASControllerStatus.csv":
                    nas_controller_status_path = os.path.join(p, f)
                elif f == "main_Main_FlightModeManagerStatus.csv":
                    fmm_status_path = os.path.join(p, f)
                elif f == "main_Boardcore_NASState.csv":
                    nas_state_path = os.path.join(p, f)
                elif f == "main_Boardcore_ADAState.csv":
                    ada_state_path = os.path.join(p, f)
            elif origin == "payload":
                if f == "payload_Payload_NASControllerStatus.csv":
                    nas_controller_status_path = os.path.join(p, f)
                elif f == "payload_Payload_FlightModeManagerStatus.csv":
                    fmm_status_path = os.path.join(p, f)
                elif f == "payload_Boardcore_NASState.csv":
                    nas_state_path = os.path.join(p, f)
                elif f == "payload_Boardcore_UBXGPSData.csv":
                    gps_path = os.path.join(p, f)

    if nas_controller_status_path is None:
        raise FileNotFoundError("NAS controller status file not found")
    if nas_state_path is None:
        raise FileNotFoundError("NAS state file not found")
    if ada_state_path is None and origin == "main":
        raise FileNotFoundError("ADA state file not found")
    if fmm_status_path is None:
        raise FileNotFoundError("FMM status file not found")
    if origin == "payload" and gps_path is None:
        raise FileNotFoundError("GPS data file not found")

    nas_controller_status = pl.read_csv(nas_controller_status_path)
    nas_state = pl.read_csv(nas_state_path)
    if origin == "main":
        ada_state = pl.read_csv(ada_state_path, infer_schema_length=10000000)
    fmm_status = pl.read_csv(fmm_status_path)

    # sort by timestamp and extract the timestamp associated to the calibrate event and topic
    nas_controller_status = nas_controller_status.sort("timestamp")
    calibrate_tms = nas_controller_status.filter(
        pl.col("state") == NAS_CALIBRATE_STATE
    ).select("timestamp")

    nas_state = nas_state.select(
        pl.from_epoch(pl.col("timestamp"), time_unit="us"),
        "n",
        "e",
        "d",
        "vn",
        "ve",
        "vd",
    )
    if origin == "main":
        ada_state = ada_state.select(
            pl.from_epoch(pl.col("timestamp"), time_unit="us"),
            "aglAltitude",
            "verticalSpeed",
        )
    fmm_status = fmm_status.select(
        pl.from_epoch(pl.col("timestamp"), time_unit="us"), "state"
    )

    # find stop timestamp based on last state of the fmm plus 10 seconds
    stop_ts = fmm_status.select(pl.col("timestamp") + pl.duration(seconds=2))[-1, 0]

    # filter the reference values and nas state dataframes
    nas_state = nas_state.filter(pl.col("timestamp") <= stop_ts)
    if origin == "main":
        ada_state = ada_state.filter(pl.col("timestamp") <= stop_ts)

    # find max timestamp
    max_ts = nas_state.select("timestamp").max().item(0, 0)

    # upsample and downsample the dataframes
    nas_state = (
        nas_state.group_by_dynamic(pl.col("timestamp"), every="50ms")
        .agg(pl.all().last())
        .upsample(time_column="timestamp", every="50ms")
        .fill_null(strategy="forward")
    )
    if origin == "main":
        ada_state = (
            ada_state.group_by_dynamic(pl.col("timestamp"), every="50ms")
            .agg(pl.all().last())
            .upsample(time_column="timestamp", every="50ms")
            .fill_null(strategy="forward")
        )

    # filter from 15 seconds before flying
    flying_fmm_state = (
        MAIN_FLYING_FMM_STATE if origin == "main" else PAYLOAD_FLYING_FMM_STATE
    )
    start_ts = fmm_status.filter(pl.col("state") == flying_fmm_state).select(
        "timestamp"
    )[0, 0] - pl.duration(seconds=SECONDS_BEFORE_FLYING)
    nas_state = nas_state.filter(pl.col("timestamp") >= start_ts)
    if origin == "main":
        ada_state = ada_state.filter(pl.col("timestamp") >= start_ts)

    if origin == "main":
        nas_state = nas_state.with_columns(
            ada_state.select("aglAltitude", "verticalSpeed")
        )

    # save the dataframes to csv
    output = Path(output)
    if origin == "main":
        nas_state.select(
            pl.col("timestamp").dt.timestamp(time_unit="us"),
            "n",
            "e",
            "d",
            "vn",
            "ve",
            "vd",
            "aglAltitude",
            "verticalSpeed",
        ).write_csv(output / "high_rate.csv")
    elif origin == "payload":
        nas_state.select(
            pl.col("timestamp").dt.timestamp(time_unit="us"),
            "n",
            "e",
            "d",
            "vn",
            "ve",
            "vd",
        ).write_csv(output / "high_rate.csv")


if __name__ == "__main__":
    main()
